@@ -10,17 +10,18 @@ from ffcx.codegeneration.definitions import FFCXBackendDefinitions
 from ffcx.codegeneration.symbols import FFCXBackendSymbols
 
 
-class FFCXBackend(object):
+class FFCXBackend:
     """Class collecting all aspects of the FFCx backend."""
 
     def __init__(self, ir, options):
-
+        """Initialise."""
         coefficient_numbering = ir.coefficient_numbering
         coefficient_offsets = ir.coefficient_offsets
 
         original_constant_offsets = ir.original_constant_offsets
 
-        self.symbols = FFCXBackendSymbols(coefficient_numbering,
-                                          coefficient_offsets, original_constant_offsets)
-        self.definitions = FFCXBackendDefinitions(ir, self.symbols, options)
+        self.symbols = FFCXBackendSymbols(
+            coefficient_numbering, coefficient_offsets, original_constant_offsets
+        )
         self.access = FFCXBackendAccess(ir, self.symbols, options)
+        self.definitions = FFCXBackendDefinitions(ir, self.access, options)
