@@ -3,6 +3,9 @@
 # This file is part of FFCx.(https://www.fenicsproject.org)
 #
 # SPDX-License-Identifier:    LGPL-3.0-or-later
+"""Just-in-time compilation."""
+
+from __future__ import annotations
 
 import importlib
 import io
@@ -16,37 +19,55 @@ from contextlib import redirect_stdout
 from pathlib import Path
 
 import cffi
+import numpy as np
+import numpy.typing as npt
 
 import ffcx
 import ffcx.naming
+from ffcx.codegeneration.C.file_template import libraries as _libraries
 
 logger = logging.getLogger("ffcx")
 root_logger = logging.getLogger()
 
 # Get declarations directly from ufcx.h
 file_dir = os.path.dirname(os.path.abspath(__file__))
-with open(file_dir + "/ufcx.h", "r") as f:
-    ufcx_h = ''.join(f.readlines())
+with open(file_dir + "/ufcx.h") as f:
+    ufcx_h = "".join(f.readlines())
 
 header = ufcx_h.split("<HEADER_DECL>")[1].split("</HEADER_DECL>")[0].strip(" /\n")
 header = header.replace("{", "{{").replace("}", "}}")
 UFC_HEADER_DECL = header + "\n"
 
-UFC_ELEMENT_DECL = '\n'.join(re.findall('typedef struct ufcx_finite_element.*?ufcx_finite_element;', ufcx_h, re.DOTALL))
-UFC_DOFMAP_DECL = '\n'.join(re.findall('typedef struct ufcx_dofmap.*?ufcx_dofmap;', ufcx_h, re.DOTALL))
-UFC_FORM_DECL = '\n'.join(re.findall('typedef struct ufcx_form.*?ufcx_form;', ufcx_h, re.DOTALL))
+UFC_ELEMENT_DECL = "\n".join(
+    re.findall("typedef struct ufcx_finite_element.*?ufcx_finite_element;", ufcx_h, re.DOTALL)
+)
+UFC_DOFMAP_DECL = "\n".join(
+    re.findall("typedef struct ufcx_dofmap.*?ufcx_dofmap;", ufcx_h, re.DOTALL)
+)
+UFC_FORM_DECL = "\n".join(re.findall("typedef struct ufcx_form.*?ufcx_form;", ufcx_h, re.DOTALL))
 
-UFC_INTEGRAL_DECL = '\n'.join(re.findall(r'typedef void ?\(ufcx_tabulate_tensor_float32\).*?\);', ufcx_h, re.DOTALL))
-UFC_INTEGRAL_DECL += '\n'.join(re.findall(r'typedef void ?\(ufcx_tabulate_tensor_float64\).*?\);', ufcx_h, re.DOTALL))
-UFC_INTEGRAL_DECL += '\n'.join(re.findall(r'typedef void ?\(ufcx_tabulate_tensor_complex64\).*?\);', ufcx_h, re.DOTALL))
-UFC_INTEGRAL_DECL += '\n'.join(re.findall(r'typedef void ?\(ufcx_tabulate_tensor_complex128\).*?\);',
-                               ufcx_h, re.DOTALL))
-UFC_INTEGRAL_DECL += '\n'.join(re.findall(r'typedef void ?\(ufcx_tabulate_tensor_longdouble\).*?\);',
-                               ufcx_h, re.DOTALL))
+UFC_INTEGRAL_DECL = "\n".join(
+    re.findall(r"typedef void ?\(ufcx_tabulate_tensor_float32\).*?\);", ufcx_h, re.DOTALL)
+)
+UFC_INTEGRAL_DECL += "\n".join(
+    re.findall(r"typedef void ?\(ufcx_tabulate_tensor_float64\).*?\);", ufcx_h, re.DOTALL)
+)
+UFC_INTEGRAL_DECL += "\n".join(
+    re.findall(r"typedef void ?\(ufcx_tabulate_tensor_complex64\).*?\);", ufcx_h, re.DOTALL)
+)
+UFC_INTEGRAL_DECL += "\n".join(
+    re.findall(r"typedef void ?\(ufcx_tabulate_tensor_complex128\).*?\);", ufcx_h, re.DOTALL)
+)
+UFC_INTEGRAL_DECL += "\n".join(
+    re.findall(r"typedef void ?\(ufcx_tabulate_tensor_longdouble\).*?\);", ufcx_h, re.DOTALL)
+)
 
-UFC_INTEGRAL_DECL += '\n'.join(re.findall('typedef struct ufcx_integral.*?ufcx_integral;',
-                                          ufcx_h, re.DOTALL))
-UFC_EXPRESSION_DECL = '\n'.join(re.findall('typedef struct ufcx_expression.*?ufcx_expression;', ufcx_h, re.DOTALL))
+UFC_INTEGRAL_DECL += "\n".join(
+    re.findall("typedef struct ufcx_integral.*?ufcx_integral;", ufcx_h, re.DOTALL)
+)
+UFC_EXPRESSION_DECL = "\n".join(
+    re.findall("typedef struct ufcx_expression.*?ufcx_expression;", ufcx_h, re.DOTALL)
+)
 
 
 def _compute_option_signature(options):
@@ -65,12 +86,15 @@ def get_cached_module(module_name, object_names, cache_dir, timeout):
 
     try:
         # Create C file with exclusive access
-        open(c_filename, "x")
+        with open(c_filename, "x"):
+            pass
         return None, None
     except FileExistsError:
         logger.info("Cached C file already exists: " + str(c_filename))
         finder = importlib.machinery.FileFinder(
-            str(cache_dir), (importlib.machinery.ExtensionFileLoader, importlib.machinery.EXTENSION_SUFFIXES))
+            str(cache_dir),
+            (importlib.machinery.ExtensionFileLoader, importlib.machinery.EXTENSION_SUFFIXES),
+        )
         finder.invalidate_caches()
 
         # Now, wait for ready
@@ -87,8 +111,10 @@ def get_cached_module(module_name, object_names, cache_dir, timeout):
 
             logger.info(f"Waiting for {ready_name} to appear.")
             time.sleep(1)
-        raise TimeoutError(f"""JIT compilation timed out, probably due to a failed previous compile.
-        Try cleaning cache (e.g. remove {c_filename}) or increase timeout option.""")
+        raise TimeoutError(
+            "JIT compilation timed out, probably due to a failed previous compile. "
+            f"Try cleaning cache (e.g. remove {c_filename}) or increase timeout option."
+        )
 
 
 def _compilation_signature(cffi_extra_compile_args=None, cffi_debug=None):
@@ -107,15 +133,25 @@ def _compilation_signature(cffi_extra_compile_args=None, cffi_debug=None):
     )
 
 
-def compile_elements(elements, options=None, cache_dir=None, timeout=10, cffi_extra_compile_args=None,
-                     cffi_verbose=False, cffi_debug=None, cffi_libraries=None):
+def compile_elements(
+    elements,
+    options: dict[str, int | float | npt.DTypeLike] | None = None,
+    cache_dir=None,
+    timeout=10,
+    cffi_extra_compile_args=None,
+    cffi_verbose=False,
+    cffi_debug=None,
+    cffi_libraries=None,
+    visualise: bool = False,
+):
     """Compile a list of UFL elements and dofmaps into Python objects."""
     p = ffcx.options.get_options(options)
 
     # Get a signature for these elements
-    module_name = 'libffcx_elements_' + \
-        ffcx.naming.compute_signature(elements, _compute_option_signature(p)
-                                      + _compilation_signature(cffi_extra_compile_args, cffi_debug))
+    module_name = "libffcx_elements_" + ffcx.naming.compute_signature(
+        elements,
+        _compute_option_signature(p) + _compilation_signature(cffi_extra_compile_args, cffi_debug),
+    )
 
     names = []
     for e in elements:
@@ -135,15 +171,30 @@ def compile_elements(elements, options=None, cache_dir=None, timeout=10, cffi_ex
         cache_dir = Path(tempfile.mkdtemp())
 
     try:
-        decl = UFC_HEADER_DECL.format(p["scalar_type"]) + UFC_ELEMENT_DECL + UFC_DOFMAP_DECL
+        decl = (
+            UFC_HEADER_DECL.format(np.dtype(p["scalar_type"]).name)  # type: ignore
+            + UFC_ELEMENT_DECL
+            + UFC_DOFMAP_DECL
+        )
         element_template = "extern ufcx_finite_element {name};\n"
         dofmap_template = "extern ufcx_dofmap {name};\n"
         for i in range(len(elements)):
             decl += element_template.format(name=names[i * 2])
             decl += dofmap_template.format(name=names[i * 2 + 1])
 
-        impl = _compile_objects(decl, elements, names, module_name, p, cache_dir,
-                                cffi_extra_compile_args, cffi_verbose, cffi_debug, cffi_libraries)
+        impl = _compile_objects(
+            decl,
+            elements,
+            names,
+            module_name,
+            p,
+            cache_dir,
+            cffi_extra_compile_args,
+            cffi_verbose,
+            cffi_debug,
+            cffi_libraries,
+            visualise=visualise,
+        )
     except Exception as e:
         try:
             # remove c file so that it will not timeout next time
@@ -159,15 +210,25 @@ def compile_elements(elements, options=None, cache_dir=None, timeout=10, cffi_ex
     return objects, module, (decl, impl)
 
 
-def compile_forms(forms, options=None, cache_dir=None, timeout=10, cffi_extra_compile_args=None,
-                  cffi_verbose=False, cffi_debug=None, cffi_libraries=None):
+def compile_forms(
+    forms,
+    options=None,
+    cache_dir=None,
+    timeout=10,
+    cffi_extra_compile_args=None,
+    cffi_verbose=False,
+    cffi_debug=None,
+    cffi_libraries=None,
+    visualise: bool = False,
+):
     """Compile a list of UFL forms into UFC Python objects."""
     p = ffcx.options.get_options(options)
 
     # Get a signature for these forms
-    module_name = 'libffcx_forms_' + \
-        ffcx.naming.compute_signature(forms, _compute_option_signature(p)
-                                      + _compilation_signature(cffi_extra_compile_args, cffi_debug))
+    module_name = "libffcx_forms_" + ffcx.naming.compute_signature(
+        forms,
+        _compute_option_signature(p) + _compilation_signature(cffi_extra_compile_args, cffi_debug),
+    )
 
     form_names = [ffcx.naming.form_name(form, i, module_name) for i, form in enumerate(forms)]
 
@@ -180,15 +241,31 @@ def compile_forms(forms, options=None, cache_dir=None, timeout=10, cffi_extra_co
         cache_dir = Path(tempfile.mkdtemp())
 
     try:
-        decl = UFC_HEADER_DECL.format(p["scalar_type"]) + UFC_ELEMENT_DECL + UFC_DOFMAP_DECL + \
-            UFC_INTEGRAL_DECL + UFC_FORM_DECL
+        decl = (
+            UFC_HEADER_DECL.format(np.dtype(p["scalar_type"]).name)  # type: ignore
+            + UFC_ELEMENT_DECL
+            + UFC_DOFMAP_DECL
+            + UFC_INTEGRAL_DECL
+            + UFC_FORM_DECL
+        )
 
         form_template = "extern ufcx_form {name};\n"
         for name in form_names:
             decl += form_template.format(name=name)
 
-        impl = _compile_objects(decl, forms, form_names, module_name, p, cache_dir,
-                                cffi_extra_compile_args, cffi_verbose, cffi_debug, cffi_libraries)
+        impl = _compile_objects(
+            decl,
+            forms,
+            form_names,
+            module_name,
+            p,
+            cache_dir,
+            cffi_extra_compile_args,
+            cffi_verbose,
+            cffi_debug,
+            cffi_libraries,
+            visualise=visualise,
+        )
     except Exception as e:
         try:
             # remove c file so that it will not timeout next time
@@ -202,22 +279,39 @@ def compile_forms(forms, options=None, cache_dir=None, timeout=10, cffi_extra_co
     return obj, module, (decl, impl)
 
 
-def compile_expressions(expressions, options=None, cache_dir=None, timeout=10, cffi_extra_compile_args=None,
-                        cffi_verbose=False, cffi_debug=None, cffi_libraries=None):
+def compile_expressions(
+    expressions,
+    options=None,
+    cache_dir=None,
+    timeout=10,
+    cffi_extra_compile_args=None,
+    cffi_verbose: bool = False,
+    cffi_debug=None,
+    cffi_libraries=None,
+    visualise: bool = False,
+):
     """Compile a list of UFL expressions into UFC Python objects.
 
-    Options
-    ----------
-    expressions
-        List of (UFL expression, evaluation points).
-
+    Args:
+        expressions: List of (UFL expression, evaluation points).
+        options: Options
+        cache_dir: Cache directory
+        timeout: Timeout
+        cffi_extra_compile_args: Extra compilation args for CFFI
+        cffi_verbose: Use verbose compile
+        cffi_debug: Use compiler debug mode
+        cffi_libraries: libraries to use with compiler
+        visualise: Toggle visualisation
     """
     p = ffcx.options.get_options(options)
 
-    module_name = 'libffcx_expressions_' + \
-        ffcx.naming.compute_signature(expressions, _compute_option_signature(p)
-                                      + _compilation_signature(cffi_extra_compile_args, cffi_debug))
-    expr_names = [ffcx.naming.expression_name(expression, module_name) for expression in expressions]
+    module_name = "libffcx_expressions_" + ffcx.naming.compute_signature(
+        expressions,
+        _compute_option_signature(p) + _compilation_signature(cffi_extra_compile_args, cffi_debug),
+    )
+    expr_names = [
+        ffcx.naming.expression_name(expression, module_name) for expression in expressions
+    ]
 
     if cache_dir is not None:
         cache_dir = Path(cache_dir)
@@ -228,15 +322,32 @@ def compile_expressions(expressions, options=None, cache_dir=None, timeout=10, c
         cache_dir = Path(tempfile.mkdtemp())
 
     try:
-        decl = UFC_HEADER_DECL.format(p["scalar_type"]) + UFC_ELEMENT_DECL + UFC_DOFMAP_DECL + \
-            UFC_INTEGRAL_DECL + UFC_FORM_DECL + UFC_EXPRESSION_DECL
+        decl = (
+            UFC_HEADER_DECL.format(np.dtype(p["scalar_type"]).name)  # type: ignore
+            + UFC_ELEMENT_DECL
+            + UFC_DOFMAP_DECL
+            + UFC_INTEGRAL_DECL
+            + UFC_FORM_DECL
+            + UFC_EXPRESSION_DECL
+        )
 
         expression_template = "extern ufcx_expression {name};\n"
         for name in expr_names:
             decl += expression_template.format(name=name)
 
-        impl = _compile_objects(decl, expressions, expr_names, module_name, p, cache_dir,
-                                cffi_extra_compile_args, cffi_verbose, cffi_debug, cffi_libraries)
+        impl = _compile_objects(
+            decl,
+            expressions,
+            expr_names,
+            module_name,
+            p,
+            cache_dir,
+            cffi_extra_compile_args,
+            cffi_verbose,
+            cffi_debug,
+            cffi_libraries,
+            visualise=visualise,
+        )
     except Exception as e:
         try:
             # remove c file so that it will not timeout next time
@@ -250,18 +361,39 @@ def compile_expressions(expressions, options=None, cache_dir=None, timeout=10, c
     return obj, module, (decl, impl)
 
 
-def _compile_objects(decl, ufl_objects, object_names, module_name, options, cache_dir,
-                     cffi_extra_compile_args, cffi_verbose, cffi_debug, cffi_libraries):
-
+def _compile_objects(
+    decl,
+    ufl_objects,
+    object_names,
+    module_name,
+    options,
+    cache_dir,
+    cffi_extra_compile_args,
+    cffi_verbose,
+    cffi_debug,
+    cffi_libraries,
+    visualise: bool = False,
+):
     import ffcx.compiler
 
+    libraries = _libraries + cffi_libraries if cffi_libraries is not None else _libraries
+
     # JIT uses module_name as prefix, which is needed to make names of all struct/function
     # unique across modules
-    _, code_body = ffcx.compiler.compile_ufl_objects(ufl_objects, prefix=module_name, options=options)
+    _, code_body = ffcx.compiler.compile_ufl_objects(
+        ufl_objects, prefix=module_name, options=options, visualise=visualise
+    )
 
     ffibuilder = cffi.FFI()
-    ffibuilder.set_source(module_name, code_body, include_dirs=[ffcx.codegeneration.get_include_path()],
-                          extra_compile_args=cffi_extra_compile_args, libraries=cffi_libraries)
+
+    ffibuilder.set_source(
+        module_name,
+        code_body,
+        include_dirs=[ffcx.codegeneration.get_include_path()],
+        extra_compile_args=cffi_extra_compile_args,
+        libraries=libraries,
+    )
+
     ffibuilder.cdef(decl)
 
     c_filename = cache_dir.joinpath(module_name + ".c")
@@ -283,10 +415,10 @@ def _compile_objects(decl, ufl_objects, object_names, module_name, options, cach
     with redirect_stdout(f):
         ffibuilder.compile(tmpdir=cache_dir, verbose=True, debug=cffi_debug)
     s = f.getvalue()
-    if (cffi_verbose):
+    if cffi_verbose:
         print(s)
 
-    logger.info("JIT C compiler finished in {:.4f}".format(time.time() - t0))
+    logger.info(f"JIT C compiler finished in {time.time() - t0:.4f}")
 
     # Create a "status ready" file. If this fails, it is an error,
     # because it should not exist yet.
@@ -303,10 +435,11 @@ def _compile_objects(decl, ufl_objects, object_names, module_name, options, cach
 
 
 def _load_objects(cache_dir, module_name, object_names):
-
     # Create module finder that searches the compile path
     finder = importlib.machinery.FileFinder(
-        str(cache_dir), (importlib.machinery.ExtensionFileLoader, importlib.machinery.EXTENSION_SUFFIXES))
+        str(cache_dir),
+        (importlib.machinery.ExtensionFileLoader, importlib.machinery.EXTENSION_SUFFIXES),
+    )
 
     # Find module. Clear search cache to be sure dynamically created
     # (new) modules are found
